import cv2
import numpy as np
from typing import Union, Tuple, List, Sequence
from numpy import typing as npt
from itertools import chain


def get_matches(descriptors1: Union[npt.NDArray[float], Sequence[npt.NDArray[float]]], descriptors2: Union[npt.NDArray[float], Sequence[npt.NDArray[float]]], matcher: cv2.DescriptorMatcher, minmatches: int):
    if not isinstance(descriptors1, (tuple, list)):
        descriptors1 = [descriptors1]
    if not isinstance(descriptors2, (tuple, list)):
        descriptors2 = [descriptors2]
    mlength = 0
    nlength = 0
    good = []
    for d1, d2 in zip(descriptors1, descriptors2):
        matches = matcher.knnMatch(d1.astype(np.float32), d2.astype(np.float32), 2)
        for m, n in matches:
            m.queryIdx = m.queryIdx + mlength
            m.trainIdx = m.trainIdx + nlength
            n.queryIdx = n.queryIdx + nlength
            n.trainIdx = n.trainIdx + mlength
            if m.distance < 0.7 * n.distance:
                good.append(m)
        mlength += d1.shape[0]
        nlength += d2.shape[0]
    return minmatches <= len(good), good


def get_features(img: npt.NDArray[np.uint8], fdet: cv2.Feature2D, mask: npt.NDArray[np.uint8] = None) -> Tuple[List[cv2.KeyPoint], Union[npt.NDArray[np.float32], npt.NDArray[np.uint8]]]:
    """
    Given a feature detector, obtain the keypoints and descriptors found in the image.
    """
    if img.ndim > 2:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    return fdet.detectAndCompute(img, mask)


def get_keypoints_descriptors(img: npt.NDArray[np.uint8], detectors: Sequence[cv2.Feature2D], mask: npt.NDArray[np.uint8] = None) -> Tuple[Tuple[cv2.KeyPoint], Tuple[Union[npt.NDArray[np.float32], npt.NDArray[np.uint8]]]]:
    features = [get_features(img, detector, mask) for detector in detectors]
    kp = tuple(chain.from_iterable([f[0] for f in features]))
    des = tuple(f[1] for f in features if f[1] is not None)
    return kp, des


def get_match_points(kp_src: Sequence[cv2.KeyPoint], kp_dst: Sequence[cv2.KeyPoint], matches: Sequence[cv2.DMatch]) -> Tuple[npt.NDArray[float], npt.NDArray[float]]:
    """
    Converts matched keypoint objects from source and destination into numpy arrays.
    """
    assert max([m.queryIdx for m in matches]) < len(kp_src), "Match indices are larger than length of src keypoints, check order of src, dst keypoint arguments."
    assert max([m.trainIdx for m in matches]) < len(kp_dst), "Match indices are larger than length of dst keypoints, check order of src, dst keypoint arguments."
    src_pts = np.float32([kp_src[m.queryIdx].pt for m in matches]).reshape(-1, 1, 2)
    dst_pts = np.float32([kp_dst[m.trainIdx].pt for m in matches]).reshape(-1, 1, 2)
    return src_pts, dst_pts



import unittest
import numpy as np
from mosaicking import registration
import cv2


class TestRegistrationModule(unittest.TestCase):
    def setUp(self) -> None:
        self._blank_image = np.zeros((100, 100), np.uint8)
        # Create a black canvas
        self._orb_image = np.zeros((500, 500), dtype=np.uint8)
        # Draw some shapes on the image that are likely to produce ORB features
        cv2.circle(self._orb_image, (200, 200), 50, (255, 255, 255), -1)
        cv2.rectangle(self._orb_image, (300, 300), (400, 400), (255, 255, 255), -1)
        cv2.line(self._orb_image, (100, 400), (400, 100), (255, 255, 255), 3)
        self._detectors = cv2.SIFT_create(), cv2.ORB_create()

    def test_get_keypoints_descriptors(self):
        kp, des = registration.get_keypoints_descriptors(self._orb_image, self._detectors, None)
        kp_blank, des_blank = registration.get_keypoints_descriptors(self._blank_image, self._detectors, None)
        num_features = len(kp)
        num_features_blank = len(kp_blank)
        self.assertLess(num_features_blank, num_features)
        self.assertEqual(num_features_blank, 0)


if __name__ == '__main__':
    unittest.main()



import argparse

import cv2
import os
import pandas as pd
import numpy as np
import numpy.typing as npt
from pathlib import Path
from matplotlib import pyplot as plt


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("video", type=str, help="Path to video file.")
    parser.add_argument("--output_directory", type=str,
                        help="Path to directory where output mosaics are to be saved. Default is invokation path.",
                        default=".")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--start_time", type=float, help="Time (secs) to start from.")
    group.add_argument("--start_frame", type=int, help="Frame number to start from.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--finish_time", type=float, help="Time (secs) to finish at.")
    group.add_argument("--finish_frame", type=int, help="Frame number to finish at.")
    parser.add_argument("--frame_skip", type=int, default=None,
                        help="Number of frames to skip between each mosaic update.")
    parser.add_argument("--orientation_file", type=str, default=None,
                        help="Path to .csv file containing orientation measurements that transform world to the camera frame.")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--sync_points", type=float, nargs=2, default=None,
                       help="Time points (sec) where video and orientation file are in sync, used to calculate time offset between video and timestamps in orientation file.")
    group.add_argument("--time_offset", type=float, default=None,
                       help="Time offset (sec) between video and orientation file timestamps, used for synchronization.")
    parser.add_argument("--min_matches", type=int, default=4,
                        help="Minimum number of matches to proceed with registration.")
    parser.add_argument("--min_features", type=int, default=100,
                        help="Minimum number of features to detect in an image.")
    parser.add_argument("--max_warp_size", type=int, nargs='+', default=None,
                        help="Maximum size of warped image (used to prevent OOM errors), if 1 argument given then image is clipped to square, if 2 then the order is height, width.")
    parser.add_argument("--max_mosaic_size", type=int, default=None,
                        help="Largest allowable size (width or height) for mosaic. Creates a new tile if it gets bigger.")
    parser.add_argument("--save_freq", type=int, default=0,
                        help="Save frequency for output mosaic (if less than 1 then output saves at exit).")
    parser.add_argument("--scale_factor", type=float, default=0.0,
                        help="Scale the input image with constant aspect ratio.")
    parser.add_argument("--alpha", type=float, default=0.5,
                        help="Alpha blending scalar for merging new frames into mosaic.")
    parser.add_argument("--show_rotation", action="store_true",
                        help="Flag to display the rotation compensation using rotation data.")
    parser.add_argument("--show_mosaic", action="store_true", help="Flag to display the mosaic output.")
    parser.add_argument("--show_preprocessing", action="store_true", help="Flag to display the preprocessed image")
    parser.add_argument("--imadjust", action="store_true", help="Flag to preprocess image for color balance.")
    parser.add_argument("--equalize_color", action="store_true",
                        help="Flag to preprocess image for contrast equalization.")
    parser.add_argument("--equalize_luminance", action="store_true", help="Flag to preprocess image for lighting equalization.")
    parser.add_argument("-c", "--calibration", type=str, default=None,
                        help="Path to calibration file, overrides --intrinsic and --distortion.")
    parser.add_argument("-k", "--intrinsic", nargs=9, type=float, default=None,
                        help="Space delimited list of intrinsic matrix terms, Read as K[0,0],K[0,1],K[0,2],K[1,0],K[1,1],K[1,2],K[2,0],K[2,1],K[2,2]. Overriden by calibration file if intrinsic present.")
    parser.add_argument("-d", "--distortion", nargs="+", type=float, default=None,
                        help="Space delimited list of distortion coefficients, Read as K1, K2, p1, p2. Overriden by calibration file if distortion present.")
    parser.add_argument("-x", "--xrotation", type=float, default=0,
                        help="Rotation around image plane's x axis (radians).")
    parser.add_argument("-y", "--yrotation", type=float, default=0,
                        help="Rotation around image plane's y axis (radians).")
    parser.add_argument("-z", "--zrotation", type=float, default=0,
                        help="Rotation around image plane's z axis (radians).")
    parser.add_argument("-g", "--gradientclip", type=float, default=0,
                        help="Clip the gradient of severely distorted image.")
    parser.add_argument("-f", "--fisheye", action="store_true", help="Flag to use fisheye distortion model.")
    parser.add_argument("--homography", type=str, choices=["rigid", "similar", "affine", "perspective"],
                        default="similar", help="Type of 2D homography to perform.")
    group = parser.add_argument_group()
    group.add_argument("--demo", action="store_true",
                       help="Creates a video of the mosaic creation process. For demo purposes only.")
    group.add_argument("--show_demo", action="store_true", help="Display the demo while underway.")
    parser.add_argument("--features", type=str, nargs="+", choices=["ORB", "SIFT", "SURF", "BRISK", "KAZE", "ALL"],
                        default="ALL", help="Set of features to use in registration.")
    parser.add_argument("--show_matches", action="store_true", help="Display the matches.")
    parser.add_argument("--inliers_roi", action="store_true",
                        help="Only allow the convex hull of the inlier points to be displayed.")
    return parser.parse_args()


def prepare_frame(left: np.ndarray, right: np.ndarray, size: tuple):
    left_size = (int(size[0]/2), int(size[1]))
    right_size = (int(size[0]/2), int(size[1]))
    left = cv2.resize(left, left_size)
    right = cv2.resize(right, right_size)
    return np.concatenate((left, right), axis=1)


class VideoPlayer(cv2.VideoCapture):
    def __init__(self, args: argparse.Namespace):
        # if args.show:
        #     cv2.namedWindow("original", cv2.WINDOW_NORMAL)
        self._args = args
        # SETUP THE VIDEO READER
        video_path = Path(args.video).resolve()
        assert video_path.exists(), "File not found: {}".format(str(video_path))
        super().__init__(str(video_path))
        # PROPERTIES OF FRAME
        self.fps = self.get(cv2.CAP_PROP_FPS)
        self.width = int(self.get(
            cv2.CAP_PROP_FRAME_WIDTH))  # if args.scale_factor is None else int(reader.get(cv2.CAP_PROP_FRAME_WIDTH)*args.scale_factor)
        self.height = int(self.get(
            cv2.CAP_PROP_FRAME_HEIGHT))  # if args.scale_factor is None else int(reader.get(cv2.CAP_PROP_FRAME_HEIGHT)*args.scale_factor)
        self.n_frames = int(self.get(cv2.CAP_PROP_FRAME_COUNT))
        self.frame_skip = self._args.frame_skip
        self._get_starting_pos()
        print(f"Opened video {self._args.video}")

    def __iter__(self):
        return VideoPlayerIterator(self)

    def __repr__(self):
        fmt = "Frame {"+":0{}d".format(len(str(self.n_frames))) + "} "+ "of {}".format(self.n_frames)
        return fmt.format(int(self.get(cv2.CAP_PROP_POS_FRAMES)) + 1)

    def _get_starting_pos(self):
        """Set a VideoCapture object to a position (either in seconds or the frame #)"""
        if self._args.finish_time:
            self.set(cv2.CAP_PROP_POS_MSEC, self._args.finish_time * 1000.0)
            self._finish_frame = int(self.get(cv2.CAP_PROP_POS_FRAMES))
            self.set(cv2.CAP_PROP_POS_MSEC, self._args.finish_time * 1000.0)
        elif self._args.finish_frame:
            self._finish_frame = int(self._args.finish_frame - 1)
        else:
            self._finish_frame = self.n_frames - 1
        if self._args.start_time:
            self.set(cv2.CAP_PROP_POS_MSEC, self._args.start_time * 1000.0)
        elif self._args.start_frame:
            self.set(cv2.CAP_PROP_POS_FRAMES, self._args.start_frame - 1)
        else:
            self.set(cv2.CAP_PROP_POS_FRAMES, 0)
        self._start_frame = int(self.get(cv2.CAP_PROP_POS_FRAMES))
        if self._finish_frame < self._start_frame:
            raise ValueError("Finishing frame is less than starting time.")

    def evaluate_stopping(self):
        """Return true if stopping conditions met."""
        return self._finish_frame <= self.get(cv2.CAP_PROP_POS_FRAMES)

    def __len__(self):
        return len(range(*slice(self._start_frame, self._finish_frame, self.frame_skip).indices(self.n_frames)))


class VideoPlayerIterator:
    def __init__(self, videoplayer: VideoPlayer):
        self._videoplayer = videoplayer
        self._index = self._videoplayer.get(cv2.CAP_PROP_POS_FRAMES)

    def __next__(self):
        if not self._videoplayer.evaluate_stopping():
            ret, img = self._videoplayer.read()
            if self._videoplayer.frame_skip:
                if self._index + self._videoplayer.frame_skip > self._videoplayer._finish_frame:
                    self._videoplayer.set(cv2.CAP_PROP_POS_FRAMES, self._videoplayer._finish_frame)
                else:
                    self._videoplayer.set(cv2.CAP_PROP_POS_FRAMES, self._videoplayer.get(cv2.CAP_PROP_POS_FRAMES) + self._videoplayer.frame_skip)
            if ret:
                self._index = self._videoplayer.get(cv2.CAP_PROP_POS_FRAMES)
                return img
        # End of Iteration
        raise StopIteration


def check_keypress(key: str):
    return key == cv2.waitKey(1)


def get_starting_pos(cap: cv2.VideoCapture, args):
    """Set a VideoCapture object to a position (either in seconds or the frame #)"""
    if args.start_time:
        cap.set(cv2.CAP_PROP_POS_MSEC, args.start_time * 1000.0)
    elif args.start_frame:
        cap.set(cv2.CAP_PROP_POS_FRAMES, args.start_frame - 1)
    return cap


def evaluate_stopping(cap: cv2.VideoCapture, args: argparse.Namespace) -> bool:
    """Return true if stopping conditions met."""
    if args.finish_time:
        return cap.get(cv2.CAP_PROP_POS_MSEC) > args.finish_time*1000.0
    elif args.finish_frame:
        return cap.get(cv2.CAP_PROP_POS_FRAMES) > args.finish_frame-1
    return cap.get(cv2.CAP_PROP_FRAME_COUNT)-1 <= cap.get(cv2.CAP_PROP_POS_FRAMES)


def load_orientations(path: os.PathLike, args: argparse.Namespace) -> pd.DataFrame:
    """Given a path containing orientations, retrieve the orientations corresponding to a time offset between video and orientation data."""
    time_offset = args.time_offset if args.time_offset else args.sync_points[1] - args.sync_points[0]
    df = pd.read_csv(str(path), index_col="timestamp")
    df.index = df.index - time_offset
    return df[~df.duplicated()]


def plot_image_histogram(img: npt.NDArray[np.uint8], ax: plt.Axes = None) -> None:
    if ax is None:
        fig, ax = plt.subplots()
    if img.ndim < 3:
        img = img[..., None]
    colors = ('b','g','r') if img.shape[-1] == 3 else ('k',)
    for channel in range(img.shape[-1]):
        hist = cv2.calcHist([img], [channel], None, [256], [0, 256])
        ax.plot(hist, colors[channel])



import unittest

import cv2
import numpy as np
from mosaicking import preprocessing


class TestPreprocessingModule(unittest.TestCase):
    def setUp(self):
        self.img = np.random.randint(0, 256, (100, 100, 3), dtype=np.uint8)
        self.gray = cv2.cvtColor(self.img, cv2.COLOR_BGR2GRAY)

    def test_const_ar_scale(self):
        scaled_img = preprocessing.const_ar_scale(self.img, 0.5)
        self.assertEqual(scaled_img.shape, (50, 50, 3))

    def test_const_ar_scale_gray(self):
        scaled_img = preprocessing.const_ar_scale(self.gray, 0.5)
        self.assertEqual(scaled_img.shape, (50, 50))

    def test_rebalance_color(self):
        rebalanced_img = preprocessing.rebalance_color(self.img, 1.0, 1.0, 1.0)
        np.testing.assert_array_equal(rebalanced_img, self.img)

    def test_rebalance_color_gray_raises(self):
        self.assertRaises(ValueError, preprocessing.rebalance_color, self.gray, 1.0, 1.0, 1.0)

    def test_find_center(self):
        center = preprocessing.find_center(self.img)
        self.assertEqual(center, (50, 50))

    def test_find_center_gray(self):
        center = preprocessing.find_center(self.gray)
        self.assertEqual(center, (50, 50))

    def test_crop_to_valid_area(self):
        cropped_img, rect = preprocessing.crop_to_valid_area(self.img)
        self.assertEqual(cropped_img.shape, self.img.shape)
        self.assertEqual(rect, (0, 0, 100, 100))

    def test_crop_to_valid_area_gray(self):
        cropped_img, rect = preprocessing.crop_to_valid_area(self.gray)
        self.assertEqual(cropped_img.shape, self.gray.shape)
        self.assertEqual(rect, (0, 0, 100, 100))

    def test_imadjust(self):
        adjusted_img = preprocessing.imadjust(self.img, 99.0, (1.0, 0.1, 3.1))
        self.assertEqual(adjusted_img.shape, self.img.shape)

    def test_imdadjust_gray(self):
        adjusted_img = preprocessing.imadjust(self.gray, 99.0, (1.0, 0.1, 3.1))
        self.assertEqual(adjusted_img.shape, self.gray.shape)

    def test_fix_contrast(self):
        contrast_fixed_img = preprocessing.equalize_color(self.img)
        self.assertEqual(contrast_fixed_img.shape, self.img.shape)

    def test_fix_contrast_gray(self):
        contrast_fixed_img = preprocessing.equalize_color(self.gray)
        self.assertEqual(contrast_fixed_img.shape, self.gray.shape)

    def test_fix_light(self):
        light_fixed_img = preprocessing.equalize_luminance(self.img)
        self.assertEqual(light_fixed_img.shape, self.img.shape)

    def test_fix_light_gray(self):
        light_fixed_img = preprocessing.equalize_luminance(self.gray)
        self.assertEqual(light_fixed_img.shape, self.gray.shape)


if __name__ == '__main__':
    unittest.main()




import cv2
from cv2 import fisheye
from pathlib import Path
import sys
import numpy as np
from scipy.interpolate import interp1d
from scipy.spatial.transform import Rotation
import warnings
import traceback

import mosaicking.transformations
from mosaicking import preprocessing, utils, transformations, registration
import yaml


def main():
    args = utils.parse_args()

    reader = utils.VideoPlayer(args)

    output_path = Path(args.output_directory).resolve()
    output_path.mkdir(parents=True, exist_ok=True)

    if args.orientation_file is not None:
        ori_path = Path(args.orientation_file).resolve()
        assert ori_path.exists(), "File not found: {}".format(str(ori_path))
        if args.sync_points is None and args.time_offset is None:
            warnings.warn("No --sync_points or --time_offset argument given, assuming video and orientation file start at the same time.", UserWarning)
        orientations = utils.load_orientations(ori_path, args)
        quat_lut = interp1d(orientations.index, orientations[["qx", "qy", "qz", "qw"]], axis=0, kind='nearest')

    # IF DEMO SPECIFIED, THEN GENERATE THE DEMO VIDEO
    if args.demo:
        output_video = output_path.joinpath("demo.mp4")
        writer = cv2.VideoWriter(str(output_video), cv2.VideoWriter_fourcc(*"mp4v"), reader.fps, frameSize=(1920, 1080), isColor=True)

    # DISPLAY WINDOWS
    if args.show_mosaic:
        cv2.namedWindow(args.video, cv2.WINDOW_NORMAL)
    if args.show_rotation:
        cv2.namedWindow("ROTATION COMPENSATION", cv2.WINDOW_NORMAL)
    if args.show_preprocessing:
        cv2.namedWindow("PREPROCESSING RESULT", cv2.WINDOW_NORMAL)
    if args.show_matches:
        cv2.namedWindow("MATCHING RESULT", cv2.WINDOW_NORMAL)
    if args.show_demo:
        cv2.namedWindow("DEMO", cv2.WINDOW_NORMAL)

    # DEFINE FEATURE DETECTOR
    if "ALL" in args.features:
        models = ["ORB", "SIFT", "SURF", "BRISK", "KAZE"]
    else:
        models = args.features
    detectors = []
    for model in models:
        if model == "ORB":
            detectors.append(cv2.ORB_create())  # ORB detector is pretty good and is CC licensed
        if model == "SIFT":
            detectors.append(cv2.SIFT_create())
        if model == "SURF":
            try:
                detectors.append(cv2.xfeatures2d.SURF_create())
            except cv2.error:
                warnings.warn("Trying to use non-free SURF on OpenCV built with non-free option disabled.", UserWarning)
        if model == "FAST":
            warnings.warn("FAST features are not yet implemented.", UserWarning)
            #detectors.append(cv2.FastFeatureDetector_create())
        if model == "BRIEF":
            warnings.warn("BRIEF features are not yet implemented.", UserWarning)
            #detectors.append(cv2.xfeatures2d.BriefDescriptorExtractor_create())
        if model == "FREAK":
            warnings.warn("FREAK features are not yet implemented.", UserWarning)
            #detectors.append(cv2.xfeatures2d.FREAK_create())
        if model == "GFTT":
            warnings.warn("GFTT features are not yet implemented.", UserWarning)
            #detectors.append(cv2.GFTTDetector_create())
        if model == "BRISK":
            detectors.append(cv2.BRISK_create())
        if model == "KAZE":
            detectors.append(cv2.KAZE_create())

    # DEFINE THE MATCHER METHOD
    # FLANN parameters
    FLANN_INDEX_KDTREE = 1
    index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
    search_params = dict(checks=50)  # or pass empty dictionary
    flann = cv2.FlannBasedMatcher(index_params, search_params)

    # DEFINE THE CAMERA PROPERTIES
    # Camera Intrinsic Matrix
    # Default is set the calibration matrix to an identity matrix with transpose components centred on the image center
    K = np.eye(3)
    K[0, 2] = float(reader.width) / 2
    K[1, 2] = float(reader.height) / 2

    if args.intrinsic is not None:
        K = np.array(args.intrinsic).reshape((3, 3))  # If -k argument is defined, generate the K matrix

    if args.calibration is not None:
        # If a calibration file has been given (a ROS camera_info yaml style file)
        calibration_path = Path(args.calibration).resolve()
        assert calibration_path.exists(), "File not found: {}".format(str(calibration_path))
        with open(args.calibration, "r") as f:
            calib_data = yaml.safe_load(f)
        if 'camera_matrix' in calib_data:
            K = np.array(calib_data['camera_matrix']['data']).reshape((3, 3))
        else:
            warnings.warn(f"No camera_matrix found in {str(calibration_path)}", UserWarning)

    # Camera Lens distortion coefficients
    dist_coeff = np.zeros((4, 1), np.float64)
    if args.distortion is not None:
        dist_coeff = np.array([[d] for d in args.distortion], np.float64)
    if args.calibration is not None:
        if 'distortion_coefficients' in calib_data:
            dist_coeff = np.array(calib_data['distortion_coefficients']['data']).reshape((calib_data['distortion_coefficients']['rows'],
                                                                                          calib_data['distortion_coefficients']['cols']))
        else:
            warnings.warn(f"No distortion_coefficients found in {str(calibration_path)}", UserWarning)

    K, roi = cv2.getOptimalNewCameraMatrix(K, dist_coeff, (reader.width, reader.height), 0)

    print("K:", K)
    K_scaled = K.copy()  # A copy of K that is scaled

    # BEGIN MAIN LOOP #
    img = None
    mosaic_img = None
    init = True  # Flag to init
    nomatches = False  # Prevents consecutive tile dumps due to no match
    counter = 0  # Counter to intermittently save the output mosaic
    tile_counter = 0  # Counter for number of generated tiles.
    try:  # External try to handle any unexpected errors
        print(f"Frames available: {len(reader)}")
        print(reader)
        # Loop through until a stopping condition is reached or a frame fails to return
        for img in reader:
            print(reader)
            og = img.copy()  # keep a copy for later reference

            # Preprocess the image
            # First rectify the image
            if args.fisheye:
                img = fisheye.undistortImage(img, K, dist_coeff)
                image_mask = fisheye.undistortImage(255 * np.ones_like(img), K, dist_coeff)[:, :, 0]
            else:
                img = cv2.undistort(img, K, dist_coeff)
                image_mask = cv2.undistort(255 * np.ones_like(img), K, dist_coeff)[:, :, 0]
            # Then apply color correction if specified
            img = preprocessing.imadjust(img) if args.imadjust else img
            # Then apply contrast balancing if specified
            img = preprocessing.equalize_color(img) if args.equalize_color else img
            # Then apply light balancing if specified
            img = preprocessing.equalize_luminance(img) if args.equalize_luminance else img
            # Enhance detail
            img = preprocessing.enhance_detail(img)

            # DEBUGGING: DISPLAY THE PREPROCESSING
            if args.show_preprocessing:
                fontsize = 3
                border = np.zeros((og.shape[0], 20, 3), dtype=np.uint8)
                border[:, :, -1] = 255
                preproc_result = np.concatenate((og, border, img), axis=1)
                text_shape, baseline = cv2.getTextSize("BEFORE", cv2.FONT_HERSHEY_PLAIN, fontsize, 3)
                preproc_result = cv2.putText(preproc_result, "BEFORE", (10, 10 + text_shape[1]), cv2.FONT_HERSHEY_PLAIN,
                                             fontsize, (255, 255, 255), 3)
                text_shape, baseline = cv2.getTextSize("AFTER", cv2.FONT_HERSHEY_PLAIN, fontsize, 3)
                preproc_result = cv2.putText(preproc_result, "AFTER", (og.shape[1] + 20 + 10, 10 + text_shape[1]),
                                             cv2.FONT_HERSHEY_PLAIN,
                                             fontsize, (255, 255, 255), 3)
                cv2.imshow("PREPROCESSING RESULT", preproc_result)

            # If it's the first time, then acquire keypoints and go to next frame
            if init:
                sys.stdout.write("Initializing new mosaic.\n")
                # Detect keypoints on the first frame
                kp_prev, des_prev = registration.get_keypoints_descriptors(img, detectors)
                num_features = len(kp_prev)
                if num_features < args.min_features:
                    sys.stderr.write("Not Enough Features, Finding Good Frame.\n")
                    continue
                # Apply scaling to image if specified
                if args.scale_factor > 0.0:
                    img, kp_prev, image_mask = transformations.apply_scale(img, kp_prev, args.scale_factor,
                                                                           image_mask)
                    K_scaled = K.copy()
                    K_scaled = K_scaled * args.scale_factor
                    K_scaled[2, 2] = 1
                # Apply rotation to image if necessary
                if args.orientation_file is not None:
                    lookup_time = reader.get(cv2.CAP_PROP_POS_MSEC) / 1000.0
                    R = Rotation.from_quat(quat_lut(lookup_time))
                    # This is infinite homography
                    img, image_mask, kp_prev = transformations.apply_transform(img, K_scaled, R, np.zeros(3),
                                                                               kp_prev,
                                                                               gradient_clip=args.gradientclip,
                                                                               mask=image_mask)
                    if args.show_rotation:
                        cv2.imshow("ROTATION COMPENSATION", img)
                # If there is no orientation file specifying the world to camera transform, then rotate according to xyz arguments
                else:
                    R = Rotation.from_euler("xyz", [args.xrotation, args.yrotation, args.zrotation])
                    img, image_mask, kp_prev = transformations.apply_transform(img, K_scaled, R, np.zeros(3),
                                                                               kp_prev,
                                                                               gradient_clip=args.gradientclip,
                                                                               mask=image_mask)
                mosaic_mask = image_mask.copy()  # Initialize the mosaic mask
                mosaic_img = img.copy()  # initialize the mosaic
                prev_img = img.copy()  # store the image as previous
                A = None  # Initialize the affine aggregated homography as None so that it isn't applied straight away
                t = [0, 0]  # Initialize the translation to 0
                init = False  # Initialization complete
                sys.stdout.write("Init stage complete.\n")
                continue
            # We are now mosaicking
            else:
                # Detect keypoints in the new frame
                kp, des = registration.get_keypoints_descriptors(img, detectors)
                num_features = len(kp)
                if num_features < args.min_features:
                    sys.stderr.write("Not Enough Features, Skipping Frame.\n")
                    continue
                if args.scale_factor > 0.0:
                    img, kp, image_mask = transformations.apply_scale(img, kp, args.scale_factor, image_mask)
                    K_scaled = K.copy()
                    K_scaled = K_scaled * args.scale_factor
                    K_scaled[2, 2] = 1
                # Apply rotation to image if necessary
                if args.orientation_file is not None:
                    lookup_time = reader.get(cv2.CAP_PROP_POS_MSEC) / 1000.0
                    R = Rotation.from_quat(quat_lut(lookup_time))
                    img, image_mask, kp = transformations.apply_transform(img, K_scaled, R, np.zeros(3), kp,
                                                                          gradient_clip=args.gradientclip,
                                                                          mask=image_mask)
                    if args.show_rotation:
                        cv2.imshow("ROTATION COMPENSATION", img)
                else:
                    R = Rotation.from_euler("xyz", [args.xrotation, args.yrotation, args.zrotation])
                    img, image_mask, kp = transformations.apply_transform(img, K_scaled, R, np.zeros(3), kp,
                                                                          gradient_clip=args.gradientclip,
                                                                          mask=image_mask)
                if args.show_rotation:
                    cv2.imshow("ROTATION COMPENSATION", img)
                # Compute matches between previous frame and current frame
                ret, matches = registration.get_matches(des_prev, des, flann, args.min_matches)
                # Handle not enough matches
                if not ret:
                    # We essentially start a new mosaic
                    sys.stderr.write("Not enough matches, starting new mosaic.\n")
                    init = True
                    if not nomatches:
                        nomatches = True
                        sys.stdout.write("Cropping tile.\n")
                        tile_counter = tile_counter + 1  # Increase the tile counter
                        fpath = output_path.joinpath("tile_{:03d}.png".format(tile_counter))  # Path for output tile
                        cv2.imwrite(str(fpath), mosaic_img)  # Save the output tile
                    continue
                nomatches = False
                # Current Image Keypoints
                src_pts = np.float32([kp[m.trainIdx].pt for m in matches]).reshape(-1, 1, 2)
                if args.inliers_roi:
                    image_mask = preprocessing.convex_mask(img, src_pts)
                # Previous Image Keypoints
                dst_pts = np.float32([kp_prev[m.queryIdx].pt for m in matches]).reshape(-1, 1, 2)

                if args.show_matches:
                    img_kp = cv2.drawMatches(prev_img, kp_prev, img, kp, matches, np.zeros_like(prev_img))
                    cv2.imshow("MATCHING RESULT", img_kp)

                # Warp the previous image keypoints onto the mosaic, if the Affine homography is available.
                if A is not None:
                    A = np.concatenate((A, np.array([[0, 0, 1]])), axis=0) if A.size < 9 else A
                    dst_pts = cv2.perspectiveTransform(dst_pts, A)
                # E, _ = cv2.findEssentialMat(src_pts, dst_pts, K_scaled)
                # _, pose_R, pose_t, _ = cv2.recoverPose(E, src_pts, dst_pts, K_scaled)
                # pose_euler = Rotation.from_matrix(pose_R).as_euler("xyz", degrees=True)
                # print("Pose\n\tRotation:\n\t\tRoll: {:.2f}\tPitch: {:.2f}\tYaw: {:.2f}\n\t"
                #       "Translation:\n\t\tX: {:.2f}\tY: {:.2f}\tZ: {:.2f}".format(*pose_euler, *pose_t.squeeze()))
                # Get the new affine alignment and bounds
                # TODO Sometimes homography wraps the image around, apparently because of vanishing point and the horizon.
                #  I cannot estimate the vanishing point easily, so in this iteration we have to live with this problem.
                #  A way to address this is to mask out the components of the image that do not intersect with the ground plane.
                #  We would need to estimate the ground plane, and compute the intersection of camera rays with this plane.
                #  If they intersect, then the pixel that the ray goes through needs to be preserved, others are discarded.
                A, xbounds, ybounds = mosaicking.transformations.get_alignment(src_pts, img.shape[:2], dst_pts, mosaic_img.shape[:2],
                                                                               homography=args.homography, gradient=args.gradientclip)
                # Get the C of the top left corner of the warped image to be inserted
                t = [-min(xbounds), -min(ybounds)]

                # TODO: Handling the tiles needs an overhaul.
                #  I would like to track the coordinates of the mosaic, and assign tiles to grid on this coordinate system.
                #  The warped image should not need to be padded. Instead the warped image pixels should be assigned a
                #  coordinated within the mosaic coordinate system. Then, the tiles that share coordinates from the warped
                #  image are sequentially updated with the pixels.
                # This checks if the mosaic has reached a maximum size in either width or height dimensions.
                if args.max_mosaic_size is not None:
                    if max(xbounds) - min(xbounds) > args.max_mosaic_size or max(ybounds) - min(
                            ybounds) > args.max_mosaic_size:
                        sys.stdout.write("Max Size Reached, Cropping tile.\n")
                        tile_counter = tile_counter + 1  # Increase the tile counter
                        fpath = output_path.joinpath("tile_{:03d}.png".format(tile_counter))  # Path for output tile
                        cv2.imwrite(str(fpath), mosaic_img)  # Save the output tile
                        # First, isolate the image
                        bounds = cv2.boundingRect(cv2.cvtColor(warped, cv2.COLOR_BGR2GRAY))
                        mosaic_img = warped[bounds[1]:bounds[1] + bounds[3],
                                     bounds[0]:bounds[0] + bounds[2], :]
                        mosaic_mask = mosaic_mask[bounds[1]:bounds[1] + bounds[3],
                                      bounds[0]:bounds[0] + bounds[2]]
                        # Second, store the keypoints referenced to the new mosaic
                        points = [tuple(np.array(k.pt) - bounds[:2]) for k in kp]
                        for k, p in zip(kp, points):
                            k.pt = p
                        kp_prev = kp
                        des_prev = [d.copy() for d in des]
                        # Third, reset the affine homography
                        A = None
                        continue

                # warp the input image into the mosaic's plane
                if args.homography in ["similar", "affine"]:
                    warped = cv2.warpAffine(img, A[:2, :], (max(xbounds) - min(xbounds), max(ybounds) - min(ybounds)))
                    warped_mask = cv2.warpAffine(image_mask, A[:2, :],
                                                 (max(xbounds) - min(xbounds), max(ybounds) - min(ybounds)))
                else:
                    warped = cv2.warpPerspective(img, A, (max(xbounds) - min(xbounds), max(ybounds) - min(ybounds)))
                    warped_mask = cv2.warpPerspective(image_mask, A,
                                                      (max(xbounds) - min(xbounds), max(ybounds) - min(ybounds)))
                warped_mask = (warped_mask == 255).astype(np.uint8) * 255

                # Get the previous iteration mosaic_mask in the shape of the update
                # Get a one channel mask of that template
                mosaic_mask_ = np.zeros_like(warped_mask)
                # Insert the previous mosaic mask into the template mask
                mosaic_mask_[t[1]:mosaic_mask.shape[0] + t[1], t[0]:mosaic_mask.shape[1] + t[0]] = mosaic_mask
                # Copy the template into the mosaic placeholder
                mosaic_img_ = np.zeros_like(warped)
                # Insert the previous mosaic into the placeholder
                mosaic_img_[t[1]:mosaic_img.shape[0] + t[1], t[0]:mosaic_img.shape[1] + t[0]] = mosaic_img
                # Get the mask where mosaic and warped input intersect
                shared = cv2.bitwise_and(warped_mask, mosaic_mask_)
                # Combine the image and tile, and update
                # Insert the warped image into the mosaic placeholder
                mosaic_img = np.where(cv2.cvtColor(warped_mask, cv2.COLOR_GRAY2BGR) > 0, warped, mosaic_img_)
                # Create an alpha blend of the warped and the previous mosaic image together
                mixer = np.uint8(
                    args.alpha * mosaic_img_.astype(np.float32) + (1.0 - args.alpha) * warped.astype(np.float32))
                # Insert the blending at the intersection only
                mosaic_img = np.where(cv2.cvtColor(shared, cv2.COLOR_GRAY2BGR) > 0, mixer, mosaic_img)
                # update the tile_mask with the warped mask region
                mosaic_mask = cv2.bitwise_or(mosaic_mask_, warped_mask)

                # Display the mosaic
                if args.show_mosaic:
                    cv2.imshow(args.video, mosaic_img)

                if args.show_demo or args.demo:
                    frame = utils.prepare_frame(img, mosaic_img, (1920, 1080))
                if args.show_demo:
                    cv2.imshow("DEMO", frame)
                if args.demo:
                    writer.write(frame)

                prev_img = img.copy()  # Update the previous frame
                kp_prev = kp  # Update the previous keypoints
                des_prev = [d.copy() for d in des]  # Update the previous descriptors

                key = cv2.waitKey(1)
                counter = counter + 1
                if args.save_freq > 0 and counter % args.save_freq == 0:
                    counter = 0
                    fpath = output_path.joinpath("current_mosaic.png")
                    cv2.imwrite(str(fpath), mosaic_img)
                if key == ord("q") or key & 0xff == 27:
                    raise KeyboardInterrupt
    except KeyboardInterrupt:
        sys.stderr.write("\nUser terminated program.\n")
    except Exception:
        # Some strange error has occurred, write out to stderr, cleanup and rethrow the error
        sys.stderr.write("\nPipeline failed at frame {}\n".format(reader.get(cv2.CAP_PROP_POS_FRAMES) + 1))
        traceback.print_exc()
        fpath = output_path.joinpath("error_frame.png")
        if img is not None:
            cv2.imwrite(str(fpath), img)
        fpath = output_path.joinpath("error_mosaic.png")
        if mosaic_img is not None:
            cv2.imwrite(str(fpath), mosaic_img)
    finally:
        # Cleanup actions and exit.
        fpath = output_path.joinpath("tile_{:03d}.png".format(tile_counter))
        if mosaic_img is not None:
            cv2.imwrite(str(fpath), mosaic_img)
        cv2.destroyAllWindows()
        reader.release()
        #del reader
        if args.demo:
            writer.release()


if __name__ == "__main__":
    main()



import numpy as np
import cv2
import argparse
from pathlib import Path
import json
import sys
from sklearn.cluster import KMeans



def main():
    model_flags = {"radtan": None, "ratpoly": cv2.CALIB_RATIONAL_MODEL}
    parser = argparse.ArgumentParser()
    parser.add_argument("video", type=str, help="Path to video file.")
    parser.add_argument("calibration", type=str, help="Path to output json file containing calibration information.")
    parser.add_argument("--cb_pattern", type=int, nargs=2, default=(8, 6), help="Number of vertices (cols rows)")
    parser.add_argument("--square_size", type=float, default=30.0, help="Size of the squares in mm.")
    parser.add_argument("--reduction_fraction", type=float, default=1.0, help="Portion of samples to keep.")
    parser.add_argument("--model", type=str, default="radtan", choices=model_flags.keys(),
                        help="Choose either radial-tangent or rational polynomial models.")
    args = parser.parse_args()
    video_path = Path(args.video).resolve(True)

    reader = cv2.VideoCapture(str(video_path))
    fps = reader.get(cv2.CAP_PROP_FPS)
    width = int(reader.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(reader.get(cv2.CAP_PROP_FRAME_HEIGHT))
    nframes = int(reader.get(cv2.CAP_PROP_FRAME_COUNT))

    cols, rows = args.cb_pattern
    square_size_mm = args.square_size

    # termination criteria
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
    # prepare object points, like (0,0,0), (1,0,0), (2,0,0) ....,(6,5,0)
    objp = np.zeros((cols * rows, 3), np.float32)
    objp[:, :2] = np.mgrid[:rows, :cols].T.reshape(-1, 2)
    objp = objp * square_size_mm
    # Arrays to store object points and image points from all the images.
    objpoints = [] # 3d point in real world space
    imgpoints = np.full((nframes, cols * rows, 1, 2), np.nan, np.float32)  # 2d points in image plane.
    try:
        print("Step 1: Extracting Corners...")
        for frame in range(nframes):
            ret, img = reader.read()  # get the next frame
            if not ret:
                break
            gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)  # convert to grayscale
            # Find the chess board corners
            ret, corners = cv2.findChessboardCorners(gray, (rows, cols), cv2.CALIB_CB_FAST_CHECK | cv2.CALIB_CB_ADAPTIVE_THRESH | cv2.CALIB_CB_NORMALIZE_IMAGE | cv2.CALIB_CB_FILTER_QUADS)
            # If found, add object points, image points (after refining them)
            if ret:
                corners2 = cv2.cornerSubPix(gray, corners, (11, 11), (-1, -1), criteria)
                imgpoints[frame, ...] = corners2
            progress = frame / (nframes - 1)
            print(f"\rCorner Extraction: [{'=' * int(progress * 50):<50}] {int(progress * 100)}%", end='')
                # Draw and display the corners
                #cv2.drawChessboardCorners(img, (cols, rows), corners2, ret)
            #cv2.imshow('img', img)
            #cv2.waitKey(1)
    except Exception as err:
        print(repr(err), file=sys.stderr)
        raise err
    finally:
        reader.release()
        #cv2.destroyAllWindows()
    print(f"\rCorner Extraction: [{'=' * int(1.0 * 50):<50}] {int(1.0 * 100)}% Done!",)
    imgpoints = imgpoints[~np.isnan(imgpoints).reshape((nframes, cols*rows*2)).any(axis=1)]
    nsamples = imgpoints.shape[0]
    if args.reduction_fraction is not None and args.reduction_fraction < 1:
        print("Step 2: Clustering corners to reduce samples...", end='')
        nsamples_reduced = int(args.reduction_fraction * nsamples)
        kmeans = KMeans(nsamples_reduced, n_init='auto')
        kmeans.fit(imgpoints.reshape((nsamples, cols*rows*2)))
        labels = kmeans.predict(imgpoints.reshape((nsamples, cols*rows*2)))
        imgpoints_reduced = np.full((nsamples_reduced, cols*rows, 1, 2), np.nan, np.float32)
        for sample in range(nsamples_reduced):
            cluster_idx = np.argwhere(labels.squeeze() == sample).squeeze()
            imgpoints_reduced[sample, ...] = imgpoints[np.random.choice(cluster_idx, 1), ...]
        print(" Done!")
    else:
        print("Step 2: Skipping because reduction_fraction is not set.")
        nsamples_reduced = nsamples
        imgpoints_reduced = imgpoints.copy()
    print("Step 3: Calibrating ...", end='')
    objpoints = [objp for _ in range(nsamples_reduced)]
    imgpoints_reduced = [p.squeeze() for p in imgpoints_reduced]
    ret, K, D, rvecs, tvecs = cv2.calibrateCamera(objpoints, imgpoints_reduced, (width, height), None, None)
    print(" Done!")
    print(f"Results\nRMS re-projection error: {ret} pix.")
    print(f"Obtained Intrinsic Matrix\n{K}\nObtained Distortion Coefficients\n{D}")
    output = {"K": K.flatten().tolist(),
              "D": D.flatten().tolist()}
    out = Path(args.calibration).resolve()
    out.parent.mkdir(parents=True, exist_ok=True)
    print(f"Saving results to {out}")
    with open(args.calibration, "w") as f:
        json.dump(output, f)
    print("Have a good day.")


if __name__ == "__main__":
    main()



import cv2
import numpy as np
from numpy.typing import NDArray
from typing import Union, Tuple
from numbers import Number


# Image Pre-processing Module
# Constant aspect ratio scaling
def const_ar_scale(img: NDArray[np.uint8], scaling: float) -> NDArray[np.uint8]:
    if scaling == 1.0:
        return img
    if scaling > 1.0:
        return cv2.resize(img, (0, 0), fx=scaling, fy=scaling, interpolation=cv2.INTER_CUBIC)
    if scaling < 1.0:
        return cv2.resize(img, (0, 0), fx=scaling, fy=scaling, interpolation=cv2.INTER_AREA)


def rebalance_color(img: NDArray[np.uint8], r: float, g: float, b: float) -> NDArray[np.uint8]:
    if img.ndim != 3:
        raise ValueError(f"img should have 3 dimensions, got {img.ndim}")
    return np.clip(img.astype(float)*[b, g, r], 0, 255).astype(np.uint8)


def enhance_detail(img: NDArray[np.uint8]) -> NDArray[np.uint8]:
    """
    Apply OpenCV's detailEnhance on input image.
    """
    assert img.ndim == 3, f"img should have 3 dimensions, got {img.ndim}"
    return cv2.detailEnhance(img)


def find_center(img: np.ndarray) -> Tuple[int, int]:
    """
    Extract the centroid of the image
    """
    if img.ndim > 2:
        x, y, w, h = cv2.boundingRect(cv2.cvtColor(img, cv2.COLOR_BGR2GRAY))
    else:
        x, y, w, h = cv2.boundingRect(img)
    return int(x + w * 0.5), int(y + h * 0.5)


def convex_mask(img: np.ndarray, src_pts: np.ndarray) -> NDArray[np.uint8]:
    # Create a mask based on the convex polygon of the valid keypoints in the matched area
    image_mask = np.zeros_like(img)
    poly = cv2.convexHull(src_pts).squeeze().astype(np.int32)
    cv2.fillPoly(image_mask, pts=[poly, ], color=(255, 255, 255))
    return image_mask[:, :, 0]


def crop_to_valid_area(img: np.ndarray) -> Tuple[NDArray[np.uint8], NDArray[np.uint8]]:
    if img.ndim > 2:
        rect = cv2.boundingRect(cv2.cvtColor(img, cv2.COLOR_BGR2GRAY))
    else:
        rect = cv2.boundingRect(img)
    return img[rect[1]:rect[1]+rect[3], rect[0]:rect[0]+rect[2]], rect


def _remap_histogram(img: NDArray[np.uint8], percent: float, gamma: float) -> NDArray[np.uint8]:
    assert img.ndim == 2, f"Input image must be 2 dimensional, has {img.ndim}."
    assert 0 <= percent <= 100, f"Percentage must be between 0 and 100, got {percent}"
    quantiles = np.array([(100.0 - percent) / 200.0, 1.0 - (100.0 - percent) / 200.0])  # shave off the bottom and top percent / 2
    bounds = img.size * quantiles
    # Apply gamma correction
    img = np.clip(cv2.pow(img.astype(float) / 255.0, gamma) * 255.0, 0, 255).astype(np.uint8)
    cumhist = np.cumsum(cv2.calcHist([img], [0], None, [256], (0, 256)))  # cumulative histogram
    low_cut = np.searchsorted(cumhist, bounds[0], side="left")  # find where the cumulative histogram first exceeds the lower bounds
    high_cut = np.searchsorted(cumhist, bounds[1], side="right") - 1  # find where the cumulative histogram lastly exceeds the upper bounds
    lut = np.concatenate((
        np.zeros(low_cut, dtype=np.uint8),
        np.linspace(0, 255, high_cut - low_cut + 1, dtype=np.uint8),
        255 * np.ones(255 - high_cut, dtype=np.uint8)
    ))
    return cv2.LUT(img, lut)


# Color Correction
def imadjust(img: NDArray[np.uint8], percent: Union[float, Tuple[float, ...]] = 100.0, gamma: Union[float, Tuple[float, ...]] = 1.0) -> NDArray[np.uint8]:
    """
    Apply a similar behaviour as MATLAB's imadjust, remapping the histogram truncated by upper and lower quantiles to [0, 255]
    Params:
    @param percent: total percentage of pixels to keep, imadjust will keep the pixels within the upper and lower bounds
    defined by:
    bounds = [(100 - percent) / 200, (percent) / 200]
    Best results appear at 100 or very close to 100, smaller values result in poor color balance.
    @param gamma: apply the gamma function
    """
    if issubclass(type(percent), Number):
        percent = (percent, percent, percent)
    if issubclass(type(gamma), Number):
        gamma = (gamma, gamma, gamma)
    output = img.copy()
    if img.ndim < 3:
        return _remap_histogram(img, percent[0], gamma[0])
    for i, (channel, p, g) in enumerate(zip(cv2.split(img), percent, gamma)):
        output[..., i] = _remap_histogram(channel, p, g)
    return output.astype(np.uint8)


# Contrast Equalization
def equalize_color(img: NDArray[np.uint8], clip_limit: float = 1.0, tile_grid_size: Tuple[int, ...] = (2, 2)) -> NDArray[np.uint8]:
    """
    Given an U8 image, apply Contrast Limited Adaptive Histogram Equalization to each channel
    """
    clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=tile_grid_size)
    output = []
    for i, channel in enumerate(cv2.split(img)):
        output.append(clahe.apply(channel))
    return cv2.merge(output)


def equalize_histogram(image: NDArray[np.uint8], clip_limit: float, tile_size: tuple) -> NDArray[np.uint8]:
    assert image.dtype == np.uint8
    # Convert the image to HSV
    clahe = cv2.createCLAHE(clip_limit, tile_size)
    # Equalize the histogram of the V channel
    hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
    hsv[..., 2] = clahe.apply(hsv[..., 2])
    return cv2.cvtColor(hsv, cv2.COLOR_HSV2BGR)


# Lighting Correction
def equalize_luminance(image: NDArray[np.uint8], clip_limit: float = 3.0, tile_grid_size: Tuple[int, ...] = (7, 7)) -> NDArray[np.uint8]:
    """
    Extract the luminance (LAB space) and apply CLAHE to it.
    """
    clahe = cv2.createCLAHE(clipLimit=clip_limit, tileGridSize=tile_grid_size)
    if image.ndim < 3:
        return clahe.apply(image)
    lab = cv2.cvtColor(image, cv2.COLOR_BGR2LAB)
    lab[..., 0] = clahe.apply(lab[..., 0])
    return cv2.cvtColor(lab, cv2.COLOR_LAB2BGR)


def unsharpen_image(image: NDArray[np.uint8], kernel_size: Tuple[int, ...], sigma: float) -> NDArray[np.uint8]:
    return image - cv2.GaussianBlur(image, kernel_size, sigma) + image


def sharpen_image(image: NDArray[np.uint8]) -> NDArray[np.uint8]:
    # Define the sharpening kernel
    kernel = np.array([[0, -1, 0], [-1, 5, -1], [0, -1, 0]])

    # Apply the kernel to the image
    result = cv2.filter2D(image, -1, kernel)

    return result
